use multipart::client::lazy::Multipart;
use std::error::Error;
use std::fmt::{Display, Formatter};
use std::io;
use std::io::Read;

#[derive(Debug)]
struct DiscordJsonError {
    invalid_json: String,
}

impl Display for DiscordJsonError {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "Discord API sent invalid JSON: {}", self.invalid_json)
    }
}

impl DiscordJsonError {
    fn new(invalid_json: String) -> Self {
        Self { invalid_json }
    }
}

impl Error for DiscordJsonError {}

struct CountingRead<'a, T, F> {
    inner_stream: &'a mut T,
    current_progress: u64,
    progress_callback: F,
}

impl<'a, T: Read, F: FnMut(u64)> CountingRead<'a, T, F> {
    fn new(inner_stream: &'a mut T, progress_callback: F) -> Self {
        Self {
            inner_stream,
            current_progress: 0,
            progress_callback,
        }
    }
}

impl<T: Read, F: FnMut(u64)> Read for CountingRead<'_, T, F> {
    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
        let result = self.inner_stream.read(buf);
        if let Ok(o) = result {
            self.current_progress += o as u64;
            (self.progress_callback)(self.current_progress);
        }
        result
    }
}

pub fn upload_discord<F: FnMut(f64)>(
    filename: &str,
    data: &[u8],
    webhook: &str,
    mut percent_callback: F,
) -> anyhow::Result<String> {
    let mut cursor = io::Cursor::new(data);
    let mut mp = Multipart::new();
    let counting_reader = CountingRead::new(&mut cursor, move |bytes| {
        (percent_callback)(bytes as f64 / data.len() as f64);
    });
    mp.add_stream("file", counting_reader, Some(filename.to_string()), None);
    let mpdata = mp.prepare()?;
    let response = ureq::post(webhook)
        .set(
            "Content-Type",
            &format!("multipart/form-data; boundary={}", mpdata.boundary()),
        )
        .send(mpdata)?;
    let response_string = response.into_string()?;
    let mut rjson = json::parse(&response_string)?;
    Ok(rjson["attachments"][0]["url"]
        .take_string()
        .ok_or_else(|| DiscordJsonError::new(response_string.clone()))?)
}
