use crate::discord::upload_discord;
use anyhow::anyhow;
use glib::Sender;
use gtk::glib::clone;
use gtk::prelude::*;
use gtk::*;
use std::fs;
use std::fs::File;
use std::io::{Read, Seek, SeekFrom};
use std::path::Path;

mod discord;
const MAXBUF: u64 = 26_214_400; // 25 MiB

fn main() {
    let app = Application::builder()
        .application_id("moe.lemonsh.zdiu")
        .build();
    app.connect_activate(build_ui);
    app.run();
}

fn errormsg<P: IsA<Window>>(text: &str, parent: Option<&P>) {
    let dialog = MessageDialog::new(
        parent,
        DialogFlags::MODAL,
        MessageType::Error,
        ButtonsType::Ok,
        text,
    );
    dialog.run();
    dialog.close();
}

enum UiThreadTask {
    UpdateProgress(f64),
    Finalize(anyhow::Result<String>),
}

fn upload_file_task(
    tx: &Sender<UiThreadTask>,
    mut file: File,
    size: u64,
    name: &str,
    webhook: &str,
) -> anyhow::Result<String> {
    let mut buffer = vec![0_u8; size as usize];
    file.read_exact(&mut buffer[..])?;
    upload_discord(name, &buffer, webhook, move |fr| {
        tx.send(UiThreadTask::UpdateProgress(fr)).unwrap();
    })
}

fn lock_gui(
    file_upload_button: &FileChooserButton,
    clip_upload_button: &Button,
    webhook_control: &Button,
    progress: &ProgressBar,
) {
    file_upload_button.set_sensitive(false);
    clip_upload_button.set_sensitive(false);
    webhook_control.set_sensitive(false);
    progress.set_fraction(0.0);
    progress.set_text(Some("Loading..."));
}

fn open_limited_and_get_name(path: &Path) -> anyhow::Result<(File, u64, String)> {
    let filename = path.file_name().unwrap().to_str().unwrap();
    let mut file_handle = File::open(path)?;
    let filesize = file_handle.seek(SeekFrom::End(0))?;
    file_handle.seek(SeekFrom::Start(0))?;
    if filesize > MAXBUF {
        Err(anyhow!("File is too big."))
    } else {
        Ok((file_handle, filesize, filename.to_string()))
    }
}

fn build_ui(app: &Application) {
    let builder = gtk::Builder::from_string(include_str!("zdiu.glade"));
    let window: ApplicationWindow = builder.object("window").unwrap();
    let (tx, rx) = glib::MainContext::channel::<UiThreadTask>(glib::PRIORITY_DEFAULT);

    let file_upload_button: FileChooserButton = builder.object("file_upload_button").unwrap();
    let clip_upload_button: Button = builder.object("clip_upload_button").unwrap();
    let progress: ProgressBar = builder.object("upload_progress").unwrap();
    let copy_button: Button = builder.object("copy_button").unwrap();
    let output_field: Entry = builder.object("output_box").unwrap();
    let webhook_entry: Entry = builder.object("webhook_entry").unwrap();
    let webhook_control: Button = builder.object("webhook_control").unwrap();

    let projectdir = directories::ProjectDirs::from("moe", "lemonsh", "zdiu").unwrap();
    let configdir = projectdir.config_dir();
    if !configdir.is_dir() {
        fs::create_dir_all(configdir).unwrap();
    }
    let webhookpath = configdir.join("webhook.txt");
    if webhookpath.is_file() {
        webhook_entry.set_text(fs::read_to_string(&webhookpath).unwrap().as_str());
    }

    file_upload_button.connect_file_set(clone!(@strong tx, @weak window, @weak webhook_entry, @weak clip_upload_button, @weak webhook_control, @weak progress => move |b| {
        let webhook = webhook_entry.text().to_string();
        if webhook.is_empty() {
            errormsg("Set a webhook first.", Some(&window));
            b.unselect_all();
            return;
        }
        let file = b.filename().unwrap();
        match open_limited_and_get_name(&file) {
            Ok(o) => {
                lock_gui(b, &clip_upload_button, &webhook_control, &progress);
                let new_sender = tx.clone();
                std::thread::spawn(move || {
                    let result = upload_file_task(&new_sender, o.0, o.1, &o.2, &webhook);
                    new_sender.send(UiThreadTask::Finalize(result)).unwrap();
                });
            }
            Err(e) => {
                errormsg(e.to_string().as_str(), Some(&window));
                b.unselect_all();
            }
        };
    }));
    clip_upload_button.connect_clicked(clone!(@strong tx, @weak window, @weak webhook_entry, @weak file_upload_button, @weak webhook_control, @weak progress => move |b| {
        let webhook = webhook_entry.text().to_string();
        if webhook.is_empty() {
            errormsg("Set a webhook first.", Some(&window));
            return;
        }
        match gtk::Clipboard::get(&gdk::SELECTION_CLIPBOARD).wait_for_image() {
            Some(p) => {
                let pngbuffer = p.save_to_bufferv("png", &[]).unwrap();
                lock_gui(&file_upload_button, b, &webhook_control, &progress);
                let new_sender = tx.clone();
                std::thread::spawn(move || {
                    let result = upload_discord("zdiupload.png", &pngbuffer, webhook.as_str(), |fr| {
                        new_sender.send(UiThreadTask::UpdateProgress(fr)).unwrap();
                    });
                    new_sender.send(UiThreadTask::Finalize(result)).unwrap();
                });
            }
            None => {
                errormsg("There are no images in the clipboard.", Some(&window));
            }
        }
    }));
    copy_button.connect_clicked(clone!(@weak output_field => move |_| {
        let text = output_field.text();
        if !text.is_empty() {
            gtk::Clipboard::get(&gdk::SELECTION_CLIPBOARD).set_text(text.as_str());
        }
    }));
    webhook_control.connect_clicked(clone!(@weak window, @weak webhook_entry, @weak file_upload_button, @weak clip_upload_button => move |b| {
        if webhook_entry.get_sensitive() {
            file_upload_button.set_sensitive(true);
            clip_upload_button.set_sensitive(true);
            webhook_entry.set_sensitive(false);
            webhook_entry.set_visibility(false);
            b.set_label("Edit");
            if let Err(e) = fs::write(&webhookpath, webhook_entry.text().as_str()) {
                errormsg(format!("Couldn't save webhook: {}", e).as_str(), Some(&window));
            } else {
                println!("Saved webhook to '{:?}'", &webhookpath);
            }
        } else {
            file_upload_button.set_sensitive(false);
            clip_upload_button.set_sensitive(false);
            webhook_entry.set_sensitive(true);
            webhook_entry.set_visibility(true);
            b.set_label("Save");
        }
    }));

    rx.attach(
        None,
        clone!(@strong window => move |data| {
            match data {
                UiThreadTask::UpdateProgress(p) => {
                    progress.set_fraction(p);
                    progress.set_text(Some(format!("{:.1}%", p*100.0).as_str()));
                }
                UiThreadTask::Finalize(u) => {
                    file_upload_button.set_sensitive(true);
                    clip_upload_button.set_sensitive(true);
                    webhook_control.set_sensitive(true);
                    progress.set_fraction(0.0);
                    progress.set_text(Some("0%"));
                    match u {
                        Ok(o) => {
                            output_field.set_text(o.as_str());
                        }
                        Err(e) => {
                            errormsg(e.to_string().as_str(), Some(&window));
                        }
                    }
                }
            }
            glib::Continue(true)
        }),
    );
    window.set_application(Some(app));
    window.present();
}
